Require Import Vtac Program List Arith Even Div2 Relations String.
Require Import basis countable lang subs model basiclemmas.
Require Import FunctionalExtensionality Classical ClassicalChoice IndefiniteDescription.

Open Scope string_scope.
Open Scope list_scope.

Set Implicit Arguments.

Local Notation cid := constructive_indefinite_description.

Section trans_part1. (*********************************************************************)

Variables
  (pD : list typename)
  (pG : list (termvar * typ))
  (pty : typ)
  (names: typename -> Prop) 
  (Nci : countably_infinite names).

Definition ftn_G_ty :=
  ftn_typ pty ++ fold_right (@app _) [] (map (fun x => ftn_typ (snd x)) pG).

Definition names1 : typename -> Prop := 
  proj1_sig (cid _ (countably_infinite_split Nci)).

Definition names1r : typename -> Prop := 
  proj1_sig (cid _ 
    (proj2_sig (cid _ (countably_infinite_split Nci)))).

Lemma names_split:
  <<CI1: countably_infinite names1>> /\ <<CI1r: countably_infinite names1r>> /\
  <<UN1r: names = nunion names1 names1r>> /\ <<DJ1r: ndisj names1 names1r>>.
Proof.
  apply (proj2_sig (cid _
    (proj2_sig (cid _ (countably_infinite_split Nci))))).
Qed.

Definition names2 : typename -> Prop :=
  proj1_sig (cid _ 
    (countably_infinite_split_finite (proj1 (proj2 names_split)) ftn_G_ty)).

Definition namesnew : typename -> Prop :=
  proj1_sig (cid _ 
    (proj2_sig (cid _ 
      (countably_infinite_split_finite (proj1 (proj2 names_split)) ftn_G_ty)))).

Lemma names1r_split : 
  <<CI2: countably_infinite names2>> /\ <<CInew: countably_infinite namesnew>> /\
  <<UN2new: names1r = nunion names2 namesnew>> /\ <<DJ2new: ndisj names2 namesnew>> /\
  <<FRESH: forall a (INL: In a ftn_G_ty) (INN: namesnew a), False>>.
Proof.
  apply (proj2_sig (cid _
    (proj2_sig (cid _ 
      (countably_infinite_split_finite (proj1 (proj2 names_split)) ftn_G_ty))))).
Qed.

Variables
  (lpw1 lpw2: world_local)
  (Nam1 : wnam (wlift lpw1) <1= names1) (Nam2 : wnam (wlift lpw2) <1= names2)
  (Wih1 : winhabited (wlift lpw1)) (Wih2 : winhabited (wlift lpw2))
  (Wcs1 : wconsistent (wlift lpw1)) (Wcs2 : wconsistent (wlift lpw2)).

Definition pw1 := wlift lpw1.
Definition pw2 := wlift lpw2.

Fixpoint decomp proj ty : typ :=
  match ty with
    | Tname a => match LEM (namesnew a) with left pf => proj a | _ => Tname a end
    | Tbvar x => Tbvar x
    | Tunit => Tunit
    | Tint => Tint
    | Tbool => Tbool
    | Tprod ty1 ty2 => Tprod (decomp proj ty1) (decomp proj ty2)
    | Tsum ty1 ty2 => Tsum (decomp proj ty1) (decomp proj ty2)
    | Tarrow ty1 ty2 => Tarrow (decomp proj ty1) (decomp proj ty2)
    | Tforall ty1 => Tforall (decomp proj ty1)
    | Texists ty1 => Texists (decomp proj ty1)
    | Tmu ty1 => Tmu (decomp proj ty1)
    | Tref ty1 => Tref (decomp proj ty1)
  end.

Variables 
  (nfst nsnd : typename -> typ)
  (nfst_closed : forall a, lc_typ 0 (nfst a))
  (nsnd_closed : forall a, lc_typ 0 (nsnd a))
  (decomp_inj : forall ty1 ty2
                (EQ1 : decomp nfst ty1 = decomp nfst ty2)
                (EQ2 : decomp nsnd ty1 = decomp nsnd ty2),
              ty1 = ty2)
  (decomp_surj : forall ty1 ty2 (CL1 : lc_typ 0 ty1) (CL2 : lc_typ 0 ty2),
               exists ty, decomp nfst ty = ty1 /\ decomp nsnd ty = ty2).

Lemma Nci1 : countably_infinite names1.
Proof. apply names_split. Qed.

Lemma Nci2 : countably_infinite names2.
Proof. apply names1r_split. Qed.

Lemma Ncinew : countably_infinite namesnew.
Proof. apply names1r_split. Qed.

Lemma Nunion : forall a, names a <-> names1 a \/ names2 a \/ namesnew a.
Proof.
  assert (X:=names_split); assert (Y:=names1r_split); des.
  by rewrite UN1r, UN2new; split; intros.
Qed.

Lemma Ndisj12 : ndisj names1 names2.
Proof.
  assert (X:=names_split); assert (Y:=names1r_split); des.
  by red; intros; eapply DJ1r; eauto; rewrite UN2new; left.
Qed.

Lemma Ndisj1new : ndisj names1 namesnew.
Proof.
  assert (X:=names_split); assert (Y:=names1r_split); des.
  by red; intros; eapply DJ1r; eauto; rewrite UN2new; right.
Qed.

Lemma Ndisj2new : ndisj names2 namesnew.
Proof.
  apply names1r_split.
Qed.

Lemma NdisjTy : forall a (IN : In a (ftn_typ pty)), namesnew a -> False.
Proof.
  intros; eapply names1r_split; eauto; apply in_or_app; auto.
Qed.

Lemma NdisjGa : forall p (IN' : In p pG) a (IN : In a (ftn_typ (snd p))), namesnew a -> False.
Proof.
  intros; eapply names1r_split; eauto; apply in_or_app; right.
  clear -IN IN' H; induction pG; [by contradiction|].
  by simpls; apply in_or_app; des; subst; eauto.
Qed.

Lemma eenc_definition:
  exists EEnc : typ -> val -> val -> nat,
    forall ty ty' v1 v2 v1' v2' (EQ : EEnc ty v1 v2 = EEnc ty' v1' v2'),
    ty = ty' /\ v1 = v1' /\ v2 = v2'.
Proof.
  assert (X := cntinf_prod typ_cntinf (cntinf_prod val_cntinf val_cntinf)).
  eapply countably_infinite_set_impl_set2 in X; red in X; des.
  exists (fun ty v1 v2 => f (ty,(v1,v2))).
  by intros; eapply X in EQ; inv EQ.
Qed.

Definition EEnc := proj1_sig (cid _ eenc_definition).

Lemma EEnc_inj : 
  forall ty ty' v1 v2 v1' v2' (EQ : EEnc ty v1 v2 = EEnc ty' v1' v2'),
    ty = ty' /\ v1 = v1' /\ v2 = v2'.
Proof. apply (proj2_sig (cid _ eenc_definition)). Qed.

Lemma decomp_fst_vsubst : forall ty ty' (CL : lc_typ 0 ty') n, 
  vsubst n (decomp nfst ty') (decomp nfst ty) = decomp nfst (vsubst n ty' ty).
Proof.
  induction ty; intros; simpl;
    try (by eauto);
    try (by f_equal; auto using IHty1, IHty2);
    try (by f_equal; auto using IHty).
  - destruct (LEM (namesnew aa)) as [N'|N']. 
      by apply vsubst_id; eapply lc_mon with (n:=0); [omega|apply nfst_closed].
    by apply vsubst_id; eapply lc_mon with (n:=0); [omega|eauto].
  - by case (nat_compare bb n).
Qed.

Lemma decomp_snd_vsubst : forall ty ty' (CL : lc_typ 0 ty') n, 
  vsubst n (decomp nsnd ty') (decomp nsnd ty) = decomp nsnd (vsubst n ty' ty).
Proof.
  induction ty; intros; simpl;
    try (by eauto);
    try (by f_equal; auto using IHty1, IHty2);
    try (by f_equal; auto using IHty).
  - destruct (LEM (namesnew aa)) as [N'|N'].
      by apply vsubst_id; eapply lc_mon with (n:=0); [omega|apply nsnd_closed].
    by apply vsubst_id; eapply lc_mon with (n:=0); [omega|eauto].
  - by case (nat_compare bb n).
Qed.

Lemma decomp_lc_typ : forall ty n,
  lc_typ n (decomp nfst ty) \/ lc_typ n (decomp nsnd ty) -> lc_typ n ty.
Proof. by induction ty; simpl; intros; auto; try destruct H; des; auto. Qed.

Lemma decomp_lc_typ' : forall ty n,
  lc_typ n ty -> lc_typ n (decomp nfst ty) /\ lc_typ n (decomp nsnd ty).
Proof.
  induction ty; simpls; intros; des; hdes; eauto 10.

  case (LEM (namesnew aa)); intros; eauto.
  by split; (eapply lc_mon with (n := 0); [omega|eauto using nfst_closed, nsnd_closed]).
Qed.

Lemma decomp_tsubst_ty : forall proj d (dOK: forall x (IN: In x d), lc_typ 0 (snd x)),
  decomp proj (tsubst_typ d pty) = tsubst_typ (map_pair id (decomp proj) d) pty.
Proof.
  assert (NdisjTy := NdisjTy). (* otherwise generated IH is too weak *)
  induction pty; intros; simpls.
  - induction d; simpl.
      destruct (LEM (namesnew aa)) as [NN|NN]; auto.
      by exfalso; eapply NdisjTy; [left; reflexivity|].
    destruct a; simpl; unfold id; case (eq_typename t aa); auto.
    by intros; apply IHd; intros; apply dOK; right.
  - f_equal.
      by apply IHt1; auto; intros; eapply NdisjTy; eauto; apply in_or_app; left; eauto.
    by apply IHt2; auto; intros; eapply NdisjTy; eauto; apply in_or_app; right; eauto.
  - f_equal.
      by apply IHt1; auto; intros; eapply NdisjTy; eauto; apply in_or_app; left; eauto.
    by apply IHt2; auto; intros; eapply NdisjTy; eauto; apply in_or_app; right; eauto.
  - f_equal.
      by apply IHt1; auto; intros; eapply NdisjTy; eauto; apply in_or_app; left; eauto.
    by apply IHt2; auto; intros; eapply NdisjTy; eauto; apply in_or_app; right; eauto.
  - by f_equal; apply IHt; auto; intros; eapply NdisjTy; eauto; apply in_or_app; left; eauto.
  - by f_equal; apply IHt; auto; intros; eapply NdisjTy; eauto; apply in_or_app; left; eauto.
  - by f_equal; apply IHt; auto; intros; eapply NdisjTy; eauto; apply in_or_app; left; eauto.
  - by f_equal; apply IHt; auto; intros; eapply NdisjTy; eauto; apply in_or_app; left; eauto.
Qed.

Lemma decomp_id : forall ty (tyN : forall a, In a (ftn_typ ty) -> namesnew a -> False) proj,
  decomp proj ty = ty.
Proof.
  induction ty; intros; simpl; auto.
  - destruct (LEM (namesnew aa)) as [N|N]; auto.
    by exfalso; eapply tyN; eauto; left; eauto.
  - rewrite IHty1, IHty2; auto; intros.
      by eapply tyN; eauto; apply in_or_app; right; eauto.
    by eapply tyN; eauto; apply in_or_app; left; eauto.
  - rewrite IHty1, IHty2; auto; intros.
      by eapply tyN; eauto; apply in_or_app; right; eauto.
    by eapply tyN; eauto; apply in_or_app; left; eauto.
  - rewrite IHty1, IHty2; auto; intros.
      by eapply tyN; eauto; apply in_or_app; right; eauto.
    by eapply tyN; eauto; apply in_or_app; left; eauto.
  - by rewrite IHty; auto.
  - by rewrite IHty; auto.
  - by rewrite IHty; auto.
  - by rewrite IHty; auto.
Qed.

Lemma decomp_tsubst : forall d (dOK: forall x (IN: In x d), lc_typ 0 (snd x))
    ty (tyN : forall a, In a (ftn_typ ty) -> namesnew a -> False) proj,
  decomp proj (tsubst_typ d ty) = tsubst_typ (map_pair id (decomp proj) d) ty.
Proof.
  induction ty; intros; simpls.
  - induction d; simpl.
      destruct (LEM (namesnew aa)); auto.
      by exfalso; eauto; eapply tyN; left; reflexivity.
    destruct a; simpl; unfold id; case (eq_typename t aa); auto.
    by intros; apply IHd; intros; apply dOK; right.
  - f_equal.
      by apply IHty1; auto; intros; eapply tyN; eauto; apply in_or_app; left; eauto.
    by apply IHty2; auto; intros; eapply tyN; eauto; apply in_or_app; right; eauto.
  - f_equal.
      by apply IHty1; auto; intros; eapply tyN; eauto; apply in_or_app; left; eauto.
    by apply IHty2; auto; intros; eapply tyN; eauto; apply in_or_app; right; eauto.
  - f_equal.
      by apply IHty1; auto; intros; eapply tyN; eauto; apply in_or_app; left; eauto.
    by apply IHty2; auto; intros; eapply tyN; eauto; apply in_or_app; right; eauto.
  - by f_equal; apply IHty.
  - by f_equal; apply IHty.
  - by f_equal; apply IHty.
  - by f_equal; apply IHty.
Qed.

Inductive vrelf_fst (r: vrelf) : vrelf :=
 | vrelf_fst_intro : forall (ty ty' : flextyp)
           (E : decomp nfst ty = ty')
           (NN : forall a, names a -> ty <> FTname a)
           (NR : forall ty', ty <> FTref ty')
           v1 v3 (IN : r ty v1 v3),
         vrelf_fst r ty' v1 (Vint (EEnc ty v1 v3))
.
Hint Constructors vrelf_fst.

Inductive vrelf_snd (r: vrelf) : vrelf :=
 | vrelf_snd_intro : forall (ty ty': flextyp)
           (E : decomp nsnd ty = ty')
           (NN : forall a, names a -> ty <> FTname a)
           (NR : forall ty', ty <> FTref ty')
           v1 v3 (IN : r ty v1 v3),
         vrelf_snd r ty' (Vint (EEnc ty v1 v3)) v3
.
Hint Constructors vrelf_snd.

Lemma vrelf_fst_mon : forall r r' ty v1 v2 (REL: vrelf_fst r ty v1 v2) (LE: r <3= r'),
  vrelf_fst r' ty v1 v2.
Proof. intros; destruct REL; eauto. Qed.

Lemma vrelf_snd_mon : forall r r' ty v1 v2 (REL: vrelf_snd r ty v1 v2) (LE: r <3= r'),
  vrelf_snd r' ty v1 v2.
Proof. intros; destruct REL; eauto. Qed.

Lemma vrelf_fst_closed : forall r (CL : vrelf_closed r), vrelf_closed (vrelf_fst r).
Proof.
  red; intros; rhdes CL; unnw; assert (X := decomp_lc_typ'); hdes.
  by destruct V; rewrite <- E; eauto 10.
Qed.

Lemma vrelf_snd_closed : forall r (CL : vrelf_closed r), vrelf_closed (vrelf_snd r).
Proof.
  red; intros; rhdes CL; unnw; assert (X := decomp_lc_typ'); hdes.
  by destruct V; rewrite <- E; eauto 10.
Qed.

Definition comp2 A (r1 r2 : relation A) : relation A :=
  fun v1 v3 => exists v2, r1 v1 v2 /\ r2 v2 v3.
Definition comp3 A (r1 r2 : typ -> relation A) : typ -> relation A :=
  fun ty v1 v3 => exists v2, r1 ty v1 v2 /\ r2 ty v2 v3.
Definition bullet A (r1 r2 : typ -> relation A) : typ -> relation A :=
  fun ty v1 v3 => exists v2, r1 (decomp nfst ty) v1 v2 /\ r2 (decomp nsnd ty) v2 v3.


(* next: construction of w *)

Inductive pw1_vrel (r : vrelf) (s1 : wst pw1) (ty: flextyp) (v1 v2: val) : Prop := 
| w1_vrel_base (REL: vrelf_fst r ty v1 v2)
| w1_vrel_step r1 (REL: wval pw1 () s1 r1 ty v1 v2) (MON: r1 <3= pw1_vrel r s1).
Hint Constructors pw1_vrel.

Inductive pw2_vrel (r : vrelf) (s2 : wst pw2) (ty: flextyp) (v1 v2: val) : Prop := 
| w2_vrel_base (REL: vrelf_snd r ty v1 v2)
| w2_vrel_step r2 (REL: wval pw2 () s2 r2 ty v1 v2) (MON: r2 <3= pw2_vrel r s2).
Hint Constructors pw2_vrel.

Definition pw_val (s : wst pw1 * wst pw2) (r : vrelf) : vrelf :=
  fun ty => match ty with
   | FTname a =>
     match LEM (namesnew a) with
     | left _ => bullet (vclos (pw1_vrel r (fst s))) (vclos (pw2_vrel r (snd s))) ty
     | _ => fun _ _ => False
     end
   | _ =>   bullet (vclos (wval pw1 () (fst s) (pw1_vrel r (fst s))))
                   (vclos (wval pw2 () (snd s) (pw2_vrel r (snd s)))) ty
  end.

Definition pw_heap (s : wst pw1 * wst pw2) (r : vrelf) : hrel :=
  comp2 (wheap pw1 () (fst s) (pw1_vrel r (fst s))) (wheap pw2 () (snd s) (pw2_vrel r (snd s))).

Definition prodrel A B (r1: relation A) (r2: relation B) : relation (A*B) :=
  fun ab ab' => r1 (fst ab) (fst ab') /\ r2 (snd ab) (snd ab').

Lemma bullet_mon : forall A (r1 r1' r2 r2': typ -> relation A) ty v1 v2
    (BUL: bullet r1 r2 ty v1 v2)
    (LE1: r1 (decomp nfst ty) <2= r1' (decomp nfst ty)) 
    (LE2: r2 (decomp nsnd ty) <2= r2' (decomp nsnd ty)),
  bullet r1' r2' ty v1 v2.
Proof. red; intros; rdes BUL; eauto. Qed.

Lemma pw1_vrel_mon : forall r r' s s' ty v1 v2 
    (REL: pw1_vrel r s ty v1 v2) (LE: r <3= r') (EXT : wext _ s' s),
  pw1_vrel r' s' ty v1 v2.
Proof.
  intros; induction REL; eauto using vrelf_fst_mon, wval_mon.
Qed.

Lemma pw2_vrel_mon : forall r r' s s' ty v1 v2 
    (REL: pw2_vrel r s ty v1 v2) (LE: r <3= r') (EXT : wext _ s' s),
  pw2_vrel r' s' ty v1 v2.
Proof.
  intros; induction REL; eauto using vrelf_snd_mon, wval_mon.
Qed.

Lemma pw_val_mon: forall r r' s s' ty v1 v2 
    (REL: pw_val s r ty v1 v2) (LE : r <3= r') (EXT : prodrel (wext pw1) (wext pw2) s' s),
  pw_val s' r' ty v1 v2.
Proof.
  intros; rdes EXT; dependent destruction ty;
   try (revert REL; unfold pw_val; destruct (LEM (namesnew a)); intros; eauto);
   rrdes REL; repeat red; eauto 10 using vclos_mon, wval_mon, pw1_vrel_mon, pw2_vrel_mon.
Qed.

Lemma pw1_vrel_unfold: forall r s1 ty v1 v2 (REL: pw1_vrel r s1 ty v1 v2),
  vrelf_fst r ty v1 v2 \/ wval pw1 () s1 (pw1_vrel r s1) ty v1 v2.
Proof. by intros; induction REL; eauto using wval_mon. Qed.

Lemma pw2_vrel_unfold: forall r s2 ty v1 v2 (REL: pw2_vrel r s2 ty v1 v2),
  vrelf_snd r ty v1 v2 \/ wval pw2 () s2 (pw2_vrel r s2) ty v1 v2.
Proof. by intros; induction REL; eauto using wval_mon. Qed.

Lemma pw1_vrel_closed : forall s r (CL : vrelf_closed r), vrelf_closed (pw1_vrel r s).
Proof.
  repeat intro; induction V; [eapply vrelf_fst_closed | eapply wval_closed]; eauto.
Qed.

Lemma pw2_vrel_closed : forall s r (CL : vrelf_closed r), vrelf_closed (pw2_vrel r s).
Proof.
  repeat intro; induction V; [eapply vrelf_snd_closed | eapply wval_closed]; eauto.
Qed.

Lemma pw_val_closed : forall s r (CL : vrelf_closed r), vrelf_closed (pw_val s r).
Proof.
  red; intros; dependent destruction ty; 
    revert V; unfold pw_val; try destruct (LEM (namesnew a)); intro V; try inv V;
    destruct V as (vv & B & C); eapply vclos_closed in B; eapply vclos_closed in C;
    eauto 10 using wval_closed, pw1_vrel_closed, pw2_vrel_closed; 
    des; eauto 10 using decomp_lc_typ.
Qed.

Lemma pw_heap_mon : forall s r r' h1 h2 (REL: pw_heap s r h1 h2) (LE: r <3= r'),
  pw_heap s r' h1 h2.
Proof.
  intros; rrdes REL; repeat red.
  eauto 10 using wheap_mon, pw1_vrel_mon, pw2_vrel_mon. 
Qed.

Lemma pw_val_arrow_wf : forall s r ty' ty f1 f2
    (IN : pw_val s r (FTarrow ty' ty) f1 f2),
  funval f1 /\ funval f2.
Proof.
  intros; destruct IN as (v & IN1 & IN2).
  destruct IN1; inv REL; destruct IN2; inv REL0.
  by assert (X:=wval_arrow_wf); hdesH X; eauto.
Qed.

Lemma pw_val_forall_wf : forall s r ty f1 f2 (IN : pw_val s r (FTforall ty) f1 f2),
  tyfunval f1 /\ tyfunval f2.
Proof.
  intros; destruct IN as (v & IN1 & IN2).
  destruct IN1; inv REL; destruct IN2; inv REL0.
  by assert (X:=wval_forall_wf); hdesH X; eauto.
Qed.

Definition pw : world.
  refine 
    {| wnam := names
     ; wst := wst pw1 * wst pw2
     ; wext := prodrel (wext pw1) (wext pw2)
     ; wextpub := prodrel (wextpub pw1) (wextpub pw2)
     ; wval := fun _ => pw_val
     ; wheap := fun _ => pw_heap
     |};
  unfold prodrel in *; try by intros; des; unnw;
   eauto using wext_trans, wextpub_trans, pw_val_arrow_wf, pw_val_forall_wf.
  - by intros; eapply pw_val_mon; eauto.
  - by intros; eapply pw_val_closed; eauto.
  - by intros; eapply pw_heap_mon; eauto.
  - intros; revert WV; unfold pw_val; try destruct (LEM (namesnew a)); intros. 
    + by apply Nunion; eauto.
    + by contradict WV.
Defined.

Definition grel_fst g (s2 : wst pw2) (s1 : wst pw1) := pw1_vrel (g (s1,s2)) s1.

Definition grel_snd g (s1 : wst pw1) (s2 : wst pw2) := pw2_vrel (g (s1,s2)) s2.

Hint Unfold grel_fst grel_snd.

Lemma pw_ext: forall s1 s2 s1' s2' (EXT1: wext _ s1 s1') (EXT2: wext _ s2 s2'),
  wext pw (s1,s2) (s1',s2').
Proof. split; eauto. Qed.

Lemma grel_fst_closed : forall g (G : wglob pw g) s2 s1, vrelf_closed (grel_fst g s2 s1).
Proof.
  red; intros; rdes G; induction V; [eapply vrelf_fst_closed | eapply wval_closed]; eauto.
Qed.

Lemma grel_snd_closed : forall g (G : wglob pw g) s1 s2, vrelf_closed (grel_snd g s1 s2).
Proof.
  red; intros; rdes G; induction V; [eapply vrelf_snd_closed | eapply wval_closed]; eauto.
Qed.

Lemma grel_fst_ltfun : forall g (G : wglob pw g) s2 s1,
  ltfun (wnam pw1) (wval pw1 () s1 (grel_fst g s2 s1)) (grel_fst g s2 s1).
Proof.
  red; split; try split; unnw; intros; eauto.
  - destruct R; eauto using wval_mon.
    inv REL; destruct ty0; try (by inversion E); [by edestruct NR|].
    simpls; revert E; destruct (LEM (namesnew a)); intro.
      by edestruct NN; auto; apply Nunion; auto.
    eapply G in IN.
      by apply wnam_supports in IN.
    by edestruct NN; auto; apply Nunion; auto.
  - destruct R; eauto using wval_mon.
    inv REL; destruct ty; simpls.
    destruct (LEM (namesnew a0)).
      by edestruct NN; auto; apply Nunion; auto. 
    inversion E; subst.
    by apply Nam1 in NM; edestruct NN; auto; apply Nunion; auto. 
Qed.

Lemma grel_snd_ltfun : forall g (G : wglob pw g) s1 s2,
  ltfun (wnam pw2) (wval pw2 () s2 (grel_snd g s1 s2)) (grel_snd g s1 s2).
Proof.
  red; split; try split; unnw; intros; eauto.
  - destruct R; eauto using wval_mon.
    inv REL; destruct ty0; try (by inversion E); [by edestruct NR|].
    simpls; revert E; destruct (LEM (namesnew a)) as [N|N] _eqn : E'; intro.
      by edestruct NN; auto; apply Nunion; auto.
    eapply G in IN.
      by apply wnam_supports in IN.
    by edestruct NN; auto; apply Nunion; auto.
  - destruct R; eauto using wval_mon.
    inv REL; destruct ty; simpls.
    destruct (LEM (namesnew a0)).
      by edestruct NN; auto; apply Nunion; auto. 
    inversion E; subst.
    by apply Nam2 in NM; edestruct NN; auto; apply Nunion; auto. 
Qed.

Lemma grel_fst_wglob : forall g (G : wglob pw g) s2, wglob pw1 (grel_fst g s2).
Proof.
  by intros; red; assert (G':=G); rdes G'; split; [| split]; intros; try destruct sp;
     eauto 10 using grel_fst_closed, grel_fst_ltfun, pw1_vrel_mon, pw_ext.
Qed.

Lemma grel_snd_wglob : forall g (G : wglob pw g) s1, wglob pw2 (grel_snd g s1).
Proof.
  by intros; red; assert (G':=G); rdes G'; split; [| split]; intros; try destruct sp;
     eauto 10 using grel_snd_closed, grel_snd_ltfun, pw2_vrel_mon, pw_ext.
Qed.

Lemma bullet_w1_empty_arrow : forall g (G : wglob pw g) s1 s2 ty ty' v1 v2,
  ~ bullet (vclos (wval pw1 () s1 (grel_fst g s2 s1)))
           (vclos (vrelf_snd (g (s1, s2)))) (FTarrow ty ty') v1 v2.
Proof.
  intros; intro IN; red in IN; destruct IN as (v & IN1 & IN2).
  destruct IN1; inv REL; destruct IN2; inv REL0; inv VR0.
  edestruct wval_arrow_wf as (_ & R2); [by apply VR |].
  specialize (R2 Vunit); des; specialize (R2 hemp).
  inv R2; induction RED; induction KK; inversion H1.
    by induction KK0; inversion H0.
  by induction KK0; inversion H1.
Qed.  

Lemma bullet_w1_empty_forall : forall g (G : wglob pw g) s1 s2 ty v1 v2,
  ~ bullet (vclos (wval pw1 () s1 (grel_fst g s2 s1)))
           (vclos (vrelf_snd (g (s1, s2)))) (FTforall ty) v1 v2.
Proof.
  intros; intro IN; red in IN; destruct IN as (v & IN1 & IN2).
  destruct IN1; inv REL; destruct IN2; inv REL0; inv VR0.
  edestruct wval_forall_wf as (_ & (e2' & R2)); [by apply VR |].
  specialize (R2 hemp).
  inversion R2; induction RED; induction KK; inversion H1.
  by induction KK0; inversion H.
Qed.  

Lemma bullet_w2_empty_arrow : forall g (G : wglob pw g) s2 s1 ty ty' v1 v2,
  ~ bullet (vclos (vrelf_fst (g (s1, s2))))
           (vclos (wval pw2 () s2 (grel_snd g s1 s2))) (FTarrow ty ty') v1 v2.
Proof.
  intros; intro IN; red in IN; destruct IN as (v & IN1 & IN2).
  destruct IN1; inv REL; destruct IN2; inv REL0; inv VR.
  edestruct wval_arrow_wf as (R1 & _); [by apply VR0|].
  specialize (R1 Vunit); des.
  specialize (R1 hemp).
  inversion R1; induction RED; induction KK; inversion H1.
    by induction KK0; inversion H.
  by induction KK0; inversion H0.
Qed.  

Lemma bullet_w2_empty_forall : forall g (G : wglob pw g) s2 s1 ty v1 v2,
  ~ bullet (vclos (vrelf_fst (g (s1, s2))))
           (vclos (wval pw2 () s2 (grel_snd g s1 s2))) (FTforall ty) v1 v2.
Proof.
  intros; intro IN; red in IN; destruct IN as (v & IN1 & IN2).
  destruct IN1; inv REL; destruct IN2; inv REL0; inv VR.
  edestruct wval_forall_wf as ((e1' & R1) & _); [by apply VR0 |].
  specialize (R1 hemp).
  inversion R1; induction RED; induction KK; inversion H1.
  by induction KK0; inversion H.
Qed.  

Lemma bullet_grel_flex : forall (ty : flextyp) g (G : wglob pw g) s1 s2 v1 v2,
  bullet (vclos (grel_fst g s2 s1)) (vclos (grel_snd g s1 s2)) ty v1 v2
  <->
  vclos (g (s1, s2)) ty v1 v2.
Proof.
  intros; dependent destruction ty; split; intro IN.

  Case "(->) arrow".
  rdes IN; simpl in IN, IN0; destruct IN; inv REL; destruct IN0; inv REL0; clear REL REL0.
  destruct VR; destruct VR0.
  - inv REL; inv REL0; eapply EEnc_inj in H0; des; apply unflex_inj in H0; subst; simpl.
    by assert (X := decomp_inj _ (FTarrow ty1 ty2) E E0); apply unflex_inj in X; subst; eauto.
  - by exfalso; eapply bullet_w2_empty_arrow; eauto; red; eexists; simpl decomp; 
       eauto 10 using vclos_mon, wval_mon.
  - by exfalso; eapply bullet_w1_empty_arrow; eauto; red; eexists; simpl decomp; 
       eauto 10 using vclos_mon, wval_mon.
  - do 2 econstructor; eapply G; eexists; simpl decomp; eauto 10 using vclos_mon, wval_mon.

  Case "(<-) arrow".
  exists (Vint (EEnc (Tarrow ty1 ty2) v1 v2)); destruct IN; inv REL; split; 
   simpl; do 3 econstructor; fold (unflex (FTarrow ty1 ty2));
   constructor; vauto.

  Case "(->) ref".
  destruct IN as (v & IN1 & IN2); destruct IN1; inv REL; destruct IN2; inv REL0.
  edestruct grel_fst_wglob as (A1 & A2 & A3); eauto; eapply A3 in VR.
  edestruct grel_snd_wglob as (B1 & B2 & B3); eauto; eapply B3 in VR0.
  by do 2 econstructor; eapply G; exists v; split; simpl; eauto.

  Case "(<-) ref".
  destruct IN; inv REL; clear REL.
  edestruct G as (A1 & A2 & A3); eauto; eapply A3 in VR.
  by destruct VR as (vv & IN1 & IN2); destruct IN1; inv REL; destruct IN2; inv REL0;
     exists vv; split; simpl; do 2 econstructor; 
     [eapply grel_fst_wglob | eapply grel_snd_wglob]; eauto.

  Case "(->) forall".
  rdes IN; simpl in IN, IN0; destruct IN; inv REL; destruct IN0; inv REL0; clear REL REL0.
  destruct VR; destruct VR0.
  - do 2 econstructor; inv REL; inv REL0; eapply EEnc_inj in H0; des; apply unflex_inj in H0; subst.
    by assert (X := decomp_inj _ (FTforall ty) E E0); apply unflex_inj in X; subst.
  - by exfalso; eapply bullet_w2_empty_forall; eauto; red; eexists; simpl decomp; 
       eauto 10 using wval_mon, vclos_mon.
  - by exfalso; eapply bullet_w1_empty_forall; eauto; red; eexists; simpl decomp; 
       eauto 10 using wval_mon, vclos_mon.
  - by do 2 econstructor; eapply G; eexists; split; simpl decomp; econstructor; 
       eauto 10 using wval_mon, vclos_mon.

  Case "(<-) forall".
  exists (Vint (EEnc (Tforall ty) v1 v2)); destruct IN; inv REL; split; 
   simpl; do 3 econstructor; fold (unflex (FTforall ty));
   constructor; vauto.

  Case "(->) name".
  destruct (LEM (namesnew a)) as [N'|N'] _eqn: E'.
    by do 2 econstructor; eapply G; simpl; rewrite E'.
  destruct (LEM (names1 a)).
    exfalso; rdes IN; simpls; rewrite E' in IN, IN0.
    destruct IN; inv REL; destruct IN0; inv REL0; clear REL REL0; destruct VR0. 
    2: by apply wnam_supports in REL; simpls; apply Nam2 in REL; eapply Ndisj12; eauto.
    inv REL; destruct ty; try inversion E; revert H0; destruct (LEM (namesnew a0)) as [N0|N0] _eqn : E0.
      intro; eapply G in IN.
        by apply wnam_supports in IN; simpls; edestruct NN; eauto.
      by apply Nunion; auto.
    intro; inversion H0; subst a0; eapply G in IN; simpl in IN.
      by eapply NN; eauto; apply Nunion; eauto.
    by apply Nunion; auto.
  destruct (LEM (names2 a)).
    exfalso; rdes IN; simpls; rewrite E' in IN, IN0.
    destruct IN; inv REL; destruct IN0; inv REL0; clear REL REL0; destruct VR.
    2: by apply wnam_supports in REL; simpls; apply Nam1 in REL; eapply Ndisj12; eauto.
    inv REL; destruct ty; try inversion E; revert H0; destruct (LEM (namesnew a0)) as [N0|N0] _eqn : E0.
      intro; eapply G in IN.
        by apply wnam_supports in IN; simpls; edestruct NN; eauto.
      by apply Nunion; auto.
    intro; inversion H0; subst a0; eapply G in IN; simpl in IN.
      by eapply NN; eauto; apply Nunion; eauto.
    by apply Nunion; auto.
  destruct (LEM (names a)) as [N|N] _eqn : E.
      try (edestruct Nunion; specialize (H N); destruct H as [H|[H|H]]; contradiction).
    rdes IN; simpls; rewrite E' in IN, IN0.
    destruct IN; inv REL; destruct IN0; inv REL0; clear REL REL0; destruct VR, VR0;
      try (by apply wnam_supports in REL; simpls; eapply Nam1 in REL);
      try (by apply wnam_supports in REL0; simpls; eapply Nam2 in REL0).
    inv REL; inv REL0; eapply EEnc_inj in H0; des; apply unflex_inj in H0; subst.
    do 2 econstructor; simpls; replace (FTname a) with ty; eauto.
    by apply unflex_inj; eapply decomp_inj; simpl; rewrite E'.

  Case "(<-) name".
  destruct (LEM (namesnew a)) as [N'|N'] _eqn : E'.
    destruct IN; inv REL; edestruct G as (_ & _ & LTF); eapply LTF in VR.
      by simpl in VR; rewrite E' in VR.
    by apply Nunion; des; eauto.
  destruct (LEM (names1 a)) as [N1|N1] _eqn : E1.
    exfalso.
    destruct IN; inv REL; edestruct G as (_ & _ & LTF); eapply LTF in VR.
       by simpl in VR; rewrite E' in VR.
     by apply Nunion; des; eauto.
  destruct (LEM (names2 a)) as [N2|N2] _eqn : E2.
    exfalso.
    destruct IN; inv REL; edestruct G as (_ & _ & LTF); eapply LTF in VR.
       by simpl in VR; rewrite E' in VR.
     by apply Nunion; des; eauto.
  destruct (LEM (names a)) as [N|N] _eqn : E; 
    [by edestruct Nunion; specialize (H N); destruct H as [H|[H|H]]; contradiction|].
  by exists (Vint (EEnc (Tname a) v1 v2)); destruct IN; inv REL; split; 
     simpl; rewrite E'; econstructor; fold (unflex (FTname a));
     do 2 econstructor; constructor;
     vauto; simpl; try rewrite E'; eauto; intros ? ? X; inv X. 
Qed.

Lemma vclos_gen_lc_typ : forall
    (r : vrelf) (CL : forall ty' v1 v2, r ty' v1 v2 -> lc_typ 0 ty')
    (vc : vrel) (CL' : forall ty' v1 v2, vc ty' v1 v2 -> lc_typ 0 ty')
    ty' v1 v2 (IN : vclos_gen r vc ty' v1 v2),
  lc_typ 0 ty'.
Proof.
  by intros; inversion IN; auto; simpl; try split; try apply CL in VR; try apply CL' in VR;
    simpls; eauto; des; auto; eapply lc_vsubstD; eauto.
Qed.

Lemma vclos_gen_nonflex1 : forall r1 r2 r s1 s2
    (ty : typ) (CL : lc_typ 0 ty) (NF : ~ exists ty' : flextyp, ty = ty'),
  bullet (vclos_gen r1 s1) (vclos_gen r2 s2) ty <2= vclos_gen r (bullet s1 s2) ty.
Proof.
  intros; dependent destruction ty; simpls; destruct PR as (v & IN1 & IN2);
   try (by inv IN1; inv IN2; econstructor; try eexists; des; eauto).
  - by contradict NF; exists (FTname aa).
  - by exfalso; eapply NF; eexists (FTarrow _ _); eauto.
  - by exfalso; eapply NF; eexists (FTref _); eauto.
  - inv IN1; inv IN2.
    by econstructor; eexists; rewrite <- decomp_fst_vsubst, <- decomp_snd_vsubst; eauto.
  - by exfalso; eapply NF; eexists (FTforall _); eauto.
  - inv IN1; inv IN2; destruct decomp_surj with ty' ty'0 as (ty'' & E1 & E2); auto.
    subst; econstructor; [by instantiate (1 := ty''); apply decomp_lc_typ; eauto | ].
    by eexists; rewrite <- decomp_fst_vsubst, <- decomp_snd_vsubst; try apply decomp_lc_typ; eauto.
Qed.

Lemma vclos_gen_nonflex2 : forall r1 r2 r s1 s2
    ty (CL : lc_typ 0 ty) (NF : ~ exists ty' : flextyp, ty = ty'),
  vclos_gen r (bullet s1 s2) ty <2= bullet (vclos_gen r1 s1) (vclos_gen r2 s2) ty.
Proof.
  intros; dependent destruction ty; simpls; inv PR; try by eexists.
  - by contradict NF; exists (FTname aa).
  - by destruct VR1, VR2; des; eexists; split; simpl; apply vclos_gen_prod; eauto.
  - by destruct VR; des; eexists; split; simpl; apply vclos_gen_sum_inl; eauto;
      apply decomp_lc_typ'; eauto.
  - by destruct VR; des; eexists; split; simpl; apply vclos_gen_sum_inr; eauto;
      apply decomp_lc_typ'; eauto.
  - by exfalso; eapply NF; eexists (FTarrow _ _); eauto.
  - by exfalso; eapply NF; eexists (FTref _); eauto.
  - by destruct VR; eexists; simpl; split; apply vclos_gen_mu;
      rewrite <- decomp_fst_vsubst, <- decomp_snd_vsubst in H; simpls; des; eauto.
  - by exfalso; eapply NF; eexists (FTforall _); eauto.
  - by destruct VR; eexists; simpl; split;
      [eapply vclos_gen_exists with (decomp nfst ty')
      |eapply vclos_gen_exists with (decomp nsnd ty')
      ]; try (by apply decomp_lc_typ'; eauto); 
         by rewrite <- decomp_fst_vsubst, <- decomp_snd_vsubst in H; simpls; des; eauto.
Qed.

Lemma bullet_grel_aux1 : forall g (G : wglob pw g) s1 s2,
  vclos (grel_fst g s2 s1) <3=
      fun ty v1 v2 => forall (* (CL : lc_typ 0 ty) *) ty' (E : ty = decomp nfst ty') v3
                      (IN : vclos (grel_snd g s1 s2) (decomp nsnd ty') v2 v3),
                      vclos (g (s1, s2)) ty' v1 v3.
Proof.
  intros; simpl; eapply (vclos_induction_strong PR); clear a b c PR; intros; subst.
  destruct (classic (exists fty : flextyp, ty' = fty)) as [(fty, E')|NF]; subst.
  - apply bullet_grel_flex; auto.
    exists v2; split; try by auto.
    econstructor; eapply vclos_gen_mon; eauto.
    by intros; eapply PR.
  - econstructor; eapply vclos_gen_mon.
      eapply vclos_gen_nonflex1; eauto.
        by apply decomp_lc_typ; right; eapply vclos_closed; eauto using grel_snd_closed.
      by destruct IN; eexists v2; eauto.
    by intros; destruct PR; des; edestruct H; eauto.
Qed.

Lemma bullet_grel_aux2 : forall g (G : wglob pw g) s1 s2,
  vclos (g (s1, s2)) <3= bullet (vclos (grel_fst g s2 s1)) (vclos (grel_snd g s1 s2)).
Proof.
  intros; eapply (vclos_induction PR); clear a b c PR; intros; subst.
  destruct (classic (exists fty : flextyp, ty = fty)) as [(fty, E')|NF]; subst.
  - by eapply bullet_grel_flex; auto; dependent destruction fty; inversion IH; simpl; eauto.
  - eapply bullet_mon; try by econstructor; eauto.
    eapply vclos_gen_nonflex2; eauto.
    eapply vclos_gen_lc_typ.
    + by intros; eapply G; eauto.
    + instantiate (1 := bullet (vclos (grel_fst g s2 s1)) (vclos (grel_snd g s1 s2))).
      intros; apply decomp_lc_typ; left; destruct H.
      by eapply vclos_closed; [des | apply grel_fst_closed]; eauto.
    + by eauto.
Qed.

Lemma bullet_grel : forall ty g (G : wglob pw g) s1 s2 v1 v2,
  bullet (vclos (grel_fst g s2 s1)) (vclos (grel_snd g s1 s2)) ty v1 v2
  <->
  vclos (g (s1, s2)) ty v1 v2.
Proof.
  split; intros IN.
  - by rdes IN; eapply bullet_grel_aux1; eauto.
  - by eapply bullet_grel_aux2; eauto.
Qed.

Lemma wexp_trans : forall g (gOK: wglob pw g) s1 s2 s1' s2',
  bullet (wexp pw1 (grel_fst g s2') s1 s1') (wexp pw2 (grel_snd g s1') s2 s2') <3=
      wexp pw g (s1, s2) (s1', s2').
Proof.
  cofix.
  intros until s2'; intros tyy ee1 ee2 IN; destruct IN as (ee & IN1 & IN2); repeat red; intros.
  assert (gOK':=gOK); rdes gOK'.
  destruct HREL as (h'' & HR1 & HR2).
  eapply wexp_core_external in IN1;
    try assumption; try (by apply grel_fst_wglob); try (by apply HR1); eauto.
    2: intros; unfold grel_fst; split; intro IN. 
    2:   by destruct IN; [right; apply REL | left; eauto using wval_mon].
    2: by destruct IN; eauto.
  eapply wexp_core_external in IN2;
    try assumption; try (by apply grel_snd_wglob); try (by apply HR2); eauto.
    2: intros; unfold grel_snd; split; intro IN.
    2:   by destruct IN; [right; apply REL | left; eauto using wval_mon].
    2: by destruct IN; eauto.
  destruct IN1; destruct IN2.

  Case "div & div".
  by apply wexp_Diverge.

  Case "div & term".
  exfalso.
  by eapply reduces_val_no_diverges; [apply RED | apply DIV'].

  Case "div & call".
  exfalso.
  eapply diverges_no_stuck; try (by apply DIV'). 
  exists hh1 (app_cont K1 e1); split; unnw; eauto.
  inversion EREL; rewrite EQ1; inversion FREL; split; intros; intro R.
  - by eapply not_reduce_cont in R; auto; eapply not_reduce_app_int; auto.
  - by eapply not_value_cont; [intro | by eauto]; eapply not_value_app; eauto.
  - by eapply not_reduce_cont in R; auto; eapply not_reduce_inst_int; auto.
  - by eapply not_value_cont; [intro | by eauto]; eapply not_value_app; eauto.

  Case "term & div".
  exfalso.
  by eapply reduces_val_no_diverges; [apply RED' | apply DIV].

  Case "term & term".
  edestruct reduces_val_det; [by apply RED'| |by apply RED0| |]; eauto.
  eapply hplus_hemp_rE in PLUS0; eapply hplus_hemp_rE in PLUS2; subst.
  eapply wexp_Terminate with (s' := (s', s'0)); eauto; try edone.
  - apply bullet_grel; eauto; red. 
    by eauto 15 using vclos_mon, pw1_vrel_mon, pw2_vrel_mon, pw_ext.
  - by repeat red; eauto 15 using wheap_mon, pw1_vrel_mon, pw2_vrel_mon, pw_ext.

  Case "term & call".
  exfalso.
  eapply reduces_val_no_stuck; [by apply RED' |].
  do 2 eexists; split; [by apply RED0 |]; unnw.
  inversion EREL; rewrite EQ1; inversion FREL; split; intros; intro R.
  - by eapply not_reduce_cont in R; auto; eapply not_reduce_app_int; auto.
  - by eapply not_value_cont; [intro | by eauto]; eapply not_value_app; eauto.
  - by eapply not_reduce_cont in R; auto; eapply not_reduce_inst_int; auto.
  - by eapply not_value_cont; [intro | by eauto]; eapply not_value_app; eauto.

  Case "call & div".
  exfalso.
  eapply diverges_no_stuck; try (by apply DIV); exists hh2 (app_cont K2 e2); split; auto.
  unnw; inversion EREL; rewrite EQ2; inversion FREL; split; intros; intro R.
  - by eapply not_reduce_cont in R; auto; eapply not_reduce_app_int; auto.
  - by eapply not_value_cont; [intro | by eauto]; eapply not_value_app; eauto.
  - by eapply not_reduce_cont in R; auto; eapply not_reduce_inst_int; auto.
  - by eapply not_value_cont; [intro | by eauto]; eapply not_value_app; eauto.

  Case "call & term".
  exfalso.
  eapply reduces_val_no_stuck; [by apply RED0 |].
  do 2 eexists; split; [by apply RED' |]; unnw.
  inversion EREL; rewrite EQ2; inversion FREL; split; intros; intro R.
  - by eapply not_reduce_cont in R; auto; eapply not_reduce_app_int; auto.
  - by eapply not_value_cont; [intro | by eauto]; eapply not_value_app; eauto.
  - by eapply not_reduce_cont in R; auto; eapply not_reduce_inst_int; auto.
  - by eapply not_value_cont; [intro | by eauto]; eapply not_value_app; eauto.

  Case "call & call".

  inversion EREL; inversion EREL0; inversion FREL; inversion FREL0; subst.

  - SCase "arrow & arrow".
    edestruct reduces_stuck_det;
    [ by apply RED'
    | by intros; intro; eapply not_reduce_cont; eauto using not_reduce_app_int
    | by intro; eapply not_value_cont; [by intro V; eauto using not_value_app | by eauto]
    | by apply RED0
    | by intros; intro; eapply not_reduce_cont; eauto using not_reduce_app_int
    | by intro; eapply not_value_cont; [by intro; eauto using not_value_app | by eauto]
    | ]. 
    apply app_cont_eq in H0; des; subst; eauto using atom_app_val_val, atom_inst_val.
    inv H1; subst; eapply EEnc_inj in H3; des; subst.
    apply unflex_inj in H3; apply exp_of_val_inj in H1; subst.
    assert (FT : exists tty1 tty2, ty0 = FTarrow tty1 tty2).
      destruct ty0; simpls; eauto.
      destruct (LEM (namesnew a)) as [N'|N'] _eqn : E'.
        by exfalso; eapply NN; [apply Nunion; right |]; eauto.
      by inversion E.
    des; subst; eapply wexp_Call with (s':=(s',s'0)); eauto; try done.
    + apply hplus_hemp_rE in PLUS2; apply hplus_hemp_rE in PLUS0; subst.
      by repeat red; eauto 15 using wheap_mon, pw1_vrel_mon, pw2_vrel_mon, pw_ext.
    + eapply wstuck_App; eauto using pw_ext.
      apply bullet_grel; auto; inv E; inv E0; repeat red.
      by eauto 15 using vclos_mon, pw1_vrel_mon, pw2_vrel_mon, pw_ext.
    + intros; intros v1' v3' IN'; assert (X := g'OK); rdes X; rrdes EXTP.
      destruct s'' as (s''1, s''2); apply bullet_grel in IN'; auto.
      destruct IN' as (v2' & IN'1 & IN'2); apply wexp_trans; auto.
      inv E; inv E0; exists (app_cont K0 v2'); unfold wcont_u, wexp in *. 
      by split; eauto 10 using grel_fst_wglob, pw1_vrel_mon, grel_snd_wglob, pw2_vrel_mon,
                               pw_ext, wextpub_in_wext, wext_trans.

  - SCase "arrow & forall".
    edestruct reduces_stuck_det;
    [ by apply RED'
    | by intros; intro; eapply not_reduce_cont; eauto using not_reduce_app_int
    | by intro; eapply not_value_cont; [by intro V; eauto using not_value_app | by eauto]
    | by apply RED0
    | by intros; intro; eapply not_reduce_cont; eauto using not_reduce_inst_int
    | by intro; eapply not_value_cont; [by intro; eauto using not_value_inst | by eauto]
    | ].
    apply app_cont_eq in H0; des; subst; eauto using atom_app_val_val, atom_inst_val.
    by inversion H1.
  
  - SCase "forall & arrow".
    edestruct reduces_stuck_det;
    [ by apply RED'
    | by intros; intro; eapply not_reduce_cont; eauto using not_reduce_inst_int
    | by intro; eapply not_value_cont; [by intro V; eauto using not_value_inst | by eauto]
    | by apply RED0
    | by intros; intro; eapply not_reduce_cont; eauto using not_reduce_app_int
    | by intro; eapply not_value_cont; [by intro; eauto using not_value_app | by eauto]
    | ].
    apply app_cont_eq in H0; des; subst; eauto using atom_app_val_val, atom_inst_val.
    by inversion H1.

  - SCase "forall & forall".
    edestruct reduces_stuck_det;
    [ by apply RED'
    | by intros; intro; eapply not_reduce_cont; eauto using not_reduce_inst_int
    | by intro; eapply not_value_cont; [by intro V; eauto using not_value_inst | by eauto]
    | by apply RED0
    | by intros; intro; eapply not_reduce_cont; eauto using not_reduce_inst_int
    | by intro; eapply not_value_cont; [by intro; eauto using not_value_inst | by eauto]
    | ].
    apply app_cont_eq in H0; des; eauto using atom_app_val_val, atom_inst_val.
    inv H1; eapply EEnc_inj in H0; des; apply unflex_inj in H0; subst.
    assert (X := decomp_surj _ _ LC LC0); des; subst.
    assert (FT : exists tty, ty2 = FTforall tty).
      destruct ty2; simpls; eauto.
      destruct (LEM (namesnew a)) as [N'|N'] _eqn : E'.
        by exfalso; eapply NN; [apply Nunion; right |]; eauto.
      by inversion E.
    des; subst; eapply wexp_Call with (s':=(s',s'0)). 
    1: by eauto. 4: by eauto. 4: by eauto. 4: by eauto. 4: by eauto.
    + apply hplus_hemp_rE in PLUS2; apply hplus_hemp_rE in PLUS0; subst.
      by repeat red; eauto 15 using wheap_mon, pw1_vrel_mon, pw2_vrel_mon, pw_ext.
    + by eapply wstuck_Inst with (ty := vsubst 0 ty tty); eauto using pw_ext, decomp_lc_typ.
    + intros; intros v1' v3' IN'; assert (X := g'OK); rdes X; rrdes EXTP; inv E; inv E0.
      destruct s'' as (s''1, s''2); apply bullet_grel in IN'; auto.
      destruct IN' as (v2' & IN'1 & IN'2); apply wexp_trans; auto.
      rewrite <-decomp_fst_vsubst in *; rewrite <-decomp_snd_vsubst in *; eauto using decomp_lc_typ.
      by unfold wcont_u, wexp in *; eexists; split; 
         eauto 10 using grel_fst_wglob, pw1_vrel_mon, grel_snd_wglob, pw2_vrel_mon,
                        pw_ext, wextpub_in_wext, wext_trans.

Existential 1 := Vunit. Existential 1 := Vunit. Existential 1 := Vunit. Existential 1 := Vunit.
Existential 1 := Vunit. Existential 1 := Vunit. Existential 1 := Vunit. Existential 1 := Vunit.
Qed.

Lemma pw_inhabited : winhabited pw.
Proof.
  intros g Gg.
  edestruct Wih1 as (s1 & IN1); try by apply minvr_glob.
  edestruct Wih2 as (s2 & IN2); try by apply minvr_glob.
  exists (s1, s2); exists hemp; simpl decomp.
  by split; eapply wheap_mon; [by apply IN1 | | by apply IN2 |];
     intros; induction PR; eauto using wval_mon.
Qed.

Lemma pw_consistent : wconsistent pw.
Proof.
  red; intros; destruct s as (s1, s2); destruct APP;
    destruct FREL as (f & IN1 & IN2); destruct IN1; inv REL; destruct IN2; inv REL0.
  - apply bullet_grel in VREL; auto.
      destruct VREL as (vv & IN1' & IN2').
      edestruct Wcs1 as (e1' & e' & R1 & R & IN1''); try by (apply grel_fst_wglob; eauto).
        by eapply wstuck_App; [| | simpl decomp in VR; apply VR |]; eauto.
      edestruct Wcs2 as (e'' & e2' & R' & R2 & IN2''); try by (apply grel_snd_wglob; eauto).
        by eapply wstuck_App; [| | simpl decomp in VR0; apply VR0 |]; eauto.
      exists e1' e2'; split; auto; split; auto; apply wexp_trans; eauto.
      exists e'; assert (EE : e' = e''); try by (eapply reduce_det; [apply R | apply R']).
      by rewrite <- EE in IN2''; split.
  - assert (X := decomp_lc_typ'); hdes.
    edestruct Wcs1 as (e1' & e' & R1 & R & IN1''). 
    + by eauto using grel_fst_wglob.
    + by eapply wstuck_Inst with (ty := vsubst 0 (decomp nfst ty') (decomp nfst ty0)); eauto.
    edestruct Wcs2 as (e'' & e2' & R' & R2 & IN2'').
    + by eauto using grel_snd_wglob.
    + by eapply wstuck_Inst with (ty := vsubst 0 (decomp nsnd ty') (decomp nsnd ty0)); eauto.
    exists e1' e2'; split; auto; split; auto; apply wexp_trans; eauto.
    exists e'; assert (EE : e' = e''); try (by eapply reduce_det; [apply R | apply R']).
    by subst; rewrite <- decomp_fst_vsubst, <- decomp_snd_vsubst; auto.

  Existential 1 := hemp. Existential 1 := hemp.
Qed.

Lemma weq_trans : forall pe1 pe2 pe3
    (WEQ1 : weq pD pG (wlift lpw1) pty pe1 pe2)
    (WEQ2 : weq pD pG (wlift lpw2) pty pe2 pe3),
  weq pD pG pw pty pe1 pe3.
Proof.
  intros; cut (wequiv pD pG pw pty pe1 pe3); [by eauto 10 using pw_inhabited, pw_consistent|].

  apply wequiv_intro; try apply WEQ1; try apply WEQ2.
  red; intros; destruct s as (s1, s2); eapply wexp_trans; eauto.

  (**********************************************************************)
  (* in order to construct the mediating gammas, we use classical choice *)
  (* to do so, we first bring bullet_grel into the right form: *)
  assert (bullet_grel' :
      forall (t : typ * (val * val)),
      exists v2,
        let (ty, y) := t in let (v1, v3) := y in
        forall (IN : vclos (g (s1, s2)) ty v1 v3) (CL : lc_typ 0 ty),
          vclos (grel_fst g s2 s1) (decomp nfst ty) v1 v2 /\
          vclos (grel_snd g s1 s2) (decomp nsnd ty) v2 v3).
    intros; destruct t; destruct p.
    destruct (classic (vclos (g (s1, s2)) t v v0)) as [IN|IN]; [| by exists Vunit].
    destruct (classic (lc_typ 0 t)) as [CL|CL]; [| by exists Vunit].
    apply bullet_grel in IN; auto.
    by destruct IN as (v' & A); exists v'.

  apply choice in bullet_grel'; destruct bullet_grel' as (f & Pf).
  set (gg1 := map (fun p =>
    ( fst p,
      (fst (snd p), f (tsubst_typ d (snd (fst p)), snd p))
    )) gg).
  set (gg2 := map (fun p =>
    ( fst p,
      (f (tsubst_typ d (snd (fst p)), snd p), snd (snd p))
    )) gg).
  assert (gg1_gg2 : map_pair fst (fun x => exp_of_val (snd x)) gg1 =
                    map_pair fst (fun x => exp_of_val (fst x)) gg2).
    unfold gg1, gg2; clear WEQ1; clear WEQ2; clear ggOK.
    generalize dependent pG; induction gg; simpl; auto.
    by intros; f_equal; eapply IHgg; eauto.
  assert (gg_gg1 : map_pair fst (fun x => exp_of_val (fst x)) gg =
                   map_pair fst (fun x => exp_of_val (fst x)) gg1).
    unfold gg1, gg2; clear WEQ1; clear WEQ2; clear ggOK; clear gg1_gg2.
    generalize dependent pG; induction gg; simpl; auto.
    by intros; f_equal; eapply IHgg; eauto.
  assert (gg_gg2 : map_pair fst (fun x => exp_of_val (snd x)) gg =
                   map_pair fst (fun x => exp_of_val (snd x)) gg2).
    unfold gg1, gg2; clear WEQ1; clear WEQ2; clear ggOK; clear gg1_gg2; clear gg_gg1.
    generalize dependent pG; induction gg; simpl; auto.
    by intros; f_equal; eapply IHgg; eauto.
  (**********************************************************************)

  exists (subst_exp (map_pair fst (fun x : val * val => exp_of_val (snd x)) gg1) pe2).

  set (d1 := map_pair id (decomp nfst) d).
  set (d2 := map_pair id (decomp nsnd) d).
  assert (d1IN : forall x t1 (IN : In (x,t1) d1), exists t, t1 = decomp nfst t /\ In (x,t) d).
    clear dependent gg1; clear dependent gg2.
    clear DEQ dOK ggOK.
    induction d; intros; try by simpl; inversion IN.
    destruct IN; destruct a.
      by simpls; exists t0; inversion H; rewrite <- H1; split; auto.
    by edestruct IHd; eauto; eexists; des; split; eauto; right.
  assert (d2IN : forall x t2 (IN : In (x,t2) d2), exists t, t2 = decomp nsnd t /\ In (x,t) d).
    clear dependent gg1; clear dependent gg2.
    clear DEQ dOK ggOK d1IN.
    induction d; intros; try by simpl; inversion IN.
    destruct IN; destruct a.
      by simpls; exists t0; inversion H; rewrite <- H1; split; auto.
    by edestruct IHd; eauto; eexists; des; split; eauto; right.
  assert (E1 : decomp nfst (tsubst_typ d pty) = tsubst_typ d1 pty).
    by eapply decomp_tsubst; eauto using NdisjGa, NdisjTy.
  assert (E2 : decomp nsnd (tsubst_typ d pty) = tsubst_typ d2 pty).
    by eapply decomp_tsubst; eauto using NdisjGa, NdisjTy.

  rewrite E1, E2; assert (X:=WEQ1); rdes X; assert (Y:=WEQ2); rdes Y.
  rewrite gg1_gg2 at 2; rewrite gg_gg1, gg_gg2; split. 

  Case "w1".
  apply EQ; eauto using grel_fst_wglob.
  - by unfold d1; rewrite map_fst_map_pair, map_id.
  - intros; destruct x; apply d1IN in IN; destruct IN as (tt & E' & IN); rewrite E'.
    by apply decomp_lc_typ'; fold (snd (t, tt)); eapply dOK.
  - by rewrite <- GEQ; unfold gg1; rewrite map_map.
  - intros.
    assert (E1' : decomp nfst (tsubst_typ d ty0) = tsubst_typ d1 ty0).
      apply decomp_tsubst; intros.
        by apply dOK.
      assert (In (x, ty0) pG).
        rewrite <- GEQ; revert IN.
        clear dependent gg2; clear ggOK gg_gg1 GEQ.
        induction gg; auto; intros; destruct IN.
          by left; destruct a0 as [[? ?] [? ?]]; inversion H1.
        by right; auto.
      by eapply NdisjGa; eauto.
    rewrite <- E1'.
    assert (V : exists v3, In (x, ty0, (v1, v3)) gg /\ v2 = f (tsubst_typ d ty0, (v1, v3))).
      clear dependent gg2; clear ggOK gg_gg1 GEQ.
      induction gg; [by inversion IN |]; inversion IN.
        by inversion H; destruct a as [[] []]; simpls; inversion H1; exists v0; split; auto.
      by destruct IHgg; auto; exists x0; des; split; auto; right.
    destruct V as (? & ? & E); rewrite E.
    (* eapply Pf. TODO why does that not unify? *)
    specialize (Pf (tsubst_typ d ty0, (v1, x0))); simpl in Pf; apply Pf.
      by eapply ggOK; apply H.
    apply lc_tsubst; intros.
      by apply dOK.
    by eapply tsubst_lc; eapply vclos_closed; [|apply gOK]; eapply ggOK; eauto.

  Case "w2".
  apply EQ0; eauto using grel_snd_wglob.
  - by unfold d2; rewrite map_fst_map_pair, map_id.
  - intros; destruct x; apply d2IN in IN; destruct IN as (tt & E' & IN); rewrite E'.
    by apply decomp_lc_typ'; fold (snd (t, tt)); eapply dOK.
  - by rewrite <- GEQ; unfold gg2; rewrite map_map.
  - intros.
    assert (E2' : decomp nsnd (tsubst_typ d ty0) = tsubst_typ d2 ty0).
      apply decomp_tsubst; intros.
        by apply dOK.
      assert (In (x, ty0) pG).
        rewrite <- GEQ; revert IN.
        clear dependent gg1; clear ggOK gg_gg2 GEQ.
        induction gg; auto; intros; destruct IN.
          by left; destruct a0 as [[? ?] [? ?]]; inversion H1.
        by right; auto.
      by eapply NdisjGa; eauto.
    rewrite <- E2'.
    assert (V : exists v3, In (x, ty0, (v3, v2)) gg /\ v1 = f (tsubst_typ d ty0, (v3, v2))).
      clear dependent gg1; clear ggOK gg_gg2 GEQ.
      induction gg; [by inversion IN |]; inversion IN.
        by inversion H; destruct a as [[? ?] [? ?]]; simpls; inversion H1; exists v; split; auto.
      by destruct IHgg; auto; exists x0; des; split; auto; right.
    destruct V as (? & ? & E); rewrite E.
    specialize (Pf (tsubst_typ d ty0, (x0, v2))); simpl in Pf; apply Pf.
      by eapply ggOK; apply H.
    apply lc_tsubst; intros.
      by apply dOK.
    by eapply tsubst_lc; eapply vclos_closed; [|apply gOK]; eapply ggOK; eauto.
Qed.

End trans_part1.
